//
// Classe per rappresentare una Lista.  Doppi puntatori con "sentinella"
// Seconda versione: utilizzo di templates ed ereditarieta`.
//



#ifndef LIST_HDR
#define LIST_HDR


//////////////////////////////////////////////////////////////////////////////
// Classi base
class BaseNode {
  friend class BaseList;	// List puo` accedere alla parte privata

  BaseNode *prev;		// Puntatore all'elemento precedente
  BaseNode *next;		// Puntatore all'elemento successivo
};



typedef BaseNode* ListPos;	// La posizione nella Lista e` un puntatore
				// ad un nodo

class BaseList {
  BaseNode *sentinel;		// Puntatore alla Sentinella
  BaseNode *current;		// Puntatore all'elemento corrente
  int len;			// La lunghezza della Lista

public:
  BaseList();			// Costruttore
  virtual ~BaseList();		// Distruttore

  // Spostamenti sulla lista
  ListPos Next();		// Sposta in avanti
  ListPos Previous();		// Sposta indietro
  ListPos GoTo(ListPos p)	// Sposta alla posizione p
    { return(current = p); }
  ListPos GoToFirst()		// Posiziona sul primo elemento della Lista
    { return(current = sentinel->next); }
  ListPos GoToLast()		// Posiziona sull'ultimo elemento della Lista
    { return(current = sentinel->prev); }
  ListPos GoToEndOfList()	// Posiziona sul terminatore di Lista
    { return(current = sentinel); }

  // Test vari
  int IsEmpty()	const		// La Lista e` vuota?
    { return(sentinel->next == sentinel); }
  int IsEndOfList() const	// Sono sulla sentinella?
    { return(current == sentinel); }
  int IsFirst() const		// Sono sul primo elemento?
    { return(current == sentinel->next); }

  // Inserimento e rimozione dati
  ListPos Insert(BaseNode *d);	// Aggiunge in corrente nella Lista
  BaseNode* Delete();		// Rimuove l'elemento corrente dalla Lista (e
				// lo restituisce)

  // Varie funzioni
  ListPos Current() const	// Ritorna la posizione corrente
    { return(current); }

  int Size() const		// Ritorna la dimensione attuale della lista
    { return(len); }
};


//////////////////////////////////////////////////////////////////////////////
// Classi Derivate

// La classe Node e` una estensione della classe BaseNode, aggiungendo alla
// sua parte privata una variabile di tipo MyNType.
template<class MyNType>
class Node: private BaseNode {
  friend class List<MyNType>;	// Stack<MyNType> deve manipolare data

  MyNType data;

public:
  // Copy constructor
  Node(const MyNType &d):
    data(d)
    {}
};


// La classe List estende la classe BaseList, ridefinendo le sue funzioni in
// modo tale da poter gestire dati di qualunque tipo.
template<class MyType>
class List: public BaseList {
public:
  ~List();			// Distruttore

  // Inserimento e rimozione dati
  ListPos Insert(const MyType &d);	// Aggiunge in corrente nella Lista
  ListPos InsertFirst(const MyType &d);	// Aggiunge in testa alla Lista
  ListPos InsertLast(const MyType &d);	// Aggiunge in coda alla Lista

  MyType Delete();		// Rimuove l'elemento corrente dalla Lista (e
				// lo restituisce)

  ListPos Write(const MyType &d);// Scrive un nuovo dato nell'elemento corrente
  MyType Read() const;		// Restituisce l'elemento nella posizione
				// corrente della Lista


  // Funzioni Varie
  ListPos Clear();		// Svuota la lista senza distruggerla
};



// Svuota la lista senza distruggerla
template<class MyType>
List<MyType>::~List()
{
  Clear();
}


// Aggiunge un elemento nella Lista
template<class MyType>
ListPos List<MyType>::Insert(const MyType &d)
{
  // Creo un oggetto di tipo Node<MyType>
  Node<MyType> *nptr = new Node<MyType>(d);

  // Lo inserisco nella Lista utilizzando la classe StackBase.
  // Non servono cast.
  return(BaseList::Insert(nptr));
}



// Aggiunge un elemento all'inizio della Lista
template<class MyType>
ListPos List<MyType>::InsertFirst(const MyType &d)
{
  GoToFirst();

  return(Insert(d));
}


// Aggiunge un elemento all'inizio della Lista
template<class MyType>
ListPos List<MyType>::InsertLast(const MyType &d)
{
  GoToEndOfList();

  return(Insert(d));
}


// Restituisce l'elemento corrente della Lista
template<class MyType>
MyType List<MyType>::Read() const
{
  if (IsEndOfList()) {
    cerr << "Errore: Read(), non e` possibile leggere la sentinella!"
	 << endl;
    exit(-2);
  }

  // Usiamo la funzione della classe BaseStack per accedere all'elemento in
  // testa alla Pila.
  // Il puntatore che ci restituisce e` ad un oggetto di tipo NodeBase, ma noi
  // sappiamo che in realta` e` di tipo Node<MyType> e quindi possiamo fare
  // un cast sicuri di non avere problemi.
  Node<MyType> *npntr = (Node<MyType> *)BaseList::Current();

  return(npntr->data);
}


// Scrive un nuovo dato nell'elemento corrente
template<class MyType>
ListPos List<MyType>::Write(const MyType &d)
{
  if (IsEndOfList()) {
    cerr << "Errore: Write(), non e` possibile scrivere sulla sentinella!"
	 << endl;
    exit(-2);
  }

  BaseNode *currnt = BaseList::Current();	// Acquisisco il puntatore
						// all'elemento corrente
  ((Node<MyType> *)currnt)->data = d;		// Ci copio il nuovo elemento

  return(currnt);
}


// Rimuove l'elemento corrente della Lista (e restituisce il suo campo
// informativo).
template<class MyType>
MyType List<MyType>::Delete()
{
  Node<MyType> *npntr = (Node<MyType> *)BaseList::Delete();

  // Salvo il dato appena recuperato
  MyType tmp(npntr->data);

  // Visto che e` la classe List<> che ha allocato un oggetto, sara` questa
  // classe a deallocarlo. Notare che non e` necessario che i distruttori
  // siano virtuali visto che accedo attraverso un puntatore all'oggetto del
  // tipo giusto.
  delete npntr;

  return(tmp);
}


// Svuota la lista senza distruggerla
template<class MyType>
ListPos List<MyType>::Clear()
{
  GoToFirst();			// Vai al primo elemento della Lista

  while (!IsEndOfList())	// Finche non sono arrivato alla sentinella
    Delete();			// Cancella un elemento

  return(Current());
}

#endif
